#include <qmessagebox.h>
#include "pcrproxy.h"
#include "pcrproto.h"

pcrfreq_t CTCSS_TONE[52] = 
{0,670,693,710,719,744,770,797,825,854,885,915,948,974,1000,1035,
 1072,1109,1148,1188,1230,1273,1318,1365,1413,1462,1514,1567,1598,
 1622,1655,1679,1713,1738,1773,1799,1835,1862,1899,1928,1966,1995,
 2035,2065,2107,2181,2257,2291,2336,2418,2503,2541};


#ifdef DEBUG_VER_
#include <stdio.h>
#endif // DEBUG_VER_


PcrProxy::PcrProxy(QObject *parent, const char *name)
  : QObject( parent, name)
{

  // connect to PcrIO to send/receive messages
  connect( this, SIGNAL( sendMessage(const char *, int)),
	   &pcrIO, SLOT( sendMessageSlot(const char *, int)));
  
  connect( &pcrIO, SIGNAL( radioMessage(const char *, int)),
	   this,	SLOT( radioMessageSlot(const char *, int)) );
  
  connect( &pcrIO, SIGNAL( disconnected() ),
	   this,	SLOT( ResetRadio() ) );
 
  connect( &timerInitRadio, SIGNAL( timeout() ), 
       this,    SLOT( timeoutInitRadio() ) );

  // initialize radio conditions
  pcrinfo.freq    = DEFAULT_FREQ;
  pcrinfo.ts      = DEFAULT_TS;
  pcrinfo.mode    = DEFAULT_MODE;
  pcrinfo.vol     = 0;
  pcrinfo.squelch = 0;
  pcrinfo.filter  = DEFAULT_FILTER;
  pcrinfo.ifShift = DEFAULT_IFSHIFT;

  bzero( pcrinfo.alias, ALIASSIZE);
  bzero( pcrinfo.info, INFOSIZE);
  bzero( pcrinfo.bandScope, 256 );
  pcrinfo.bandScopeNsamples = 0x20; 
  pcrinfo.isBandScopeOn     = false;
  pcrinfo.isPowerOn         = false;

  pcrinfo.ATT   = false;
  pcrinfo.AGC   = false;
  pcrinfo.NB    = false;
  pcrinfo.VSC   = false;
  pcrinfo.AFC   = false;
  pcrinfo.CTCSS = 0;
  pcrinfo.previousCTCSS = 0;
}




PcrProxy::~PcrProxy()
{
  TurnOffRadio();
}


void PcrProxy::timeoutInitRadio()
{
  QMessageBox::critical(0,"QtPCR",
                          "QtPCR cannot initiate connection with PCR-1000.\n"
                          "Please refer to FAQ.html. Exiting ...");
}

bool PcrProxy::InitRadio(const char *device)
{
  // reset power on state
  pcrinfo.isPowerOn = false;
  
  // open serial device
  if(!pcrIO.Open(device, B9600)) return false;
  sleep(3);
  
  // turn radio on
  emit sendMessage("H101\x0d\x0aG300\x0d\x0a",12);
  sleep(3);

  // set baudrate
  emit sendMessage("G105\x0d\x0a",6);
  pcrIO.Close();
  
  // reset radio
  if(!pcrIO.Open(device, B38400)) return false;
  sleep(3);
  
  // autoupdate mode
  emit sendMessage("H101\x0d\x0aG301\x0d\x0a",12);
 
  // start counting clock
  timerInitRadio.start( 10000, true );
 
  // init radio state
  setFreq(pcrinfo.freq);
  setVol(pcrinfo.vol);
  setSquelch(pcrinfo.squelch);
  setIFShift(pcrinfo.ifShift);
  setATT(pcrinfo.ATT);
  setAGC(pcrinfo.AGC);
  setNB(pcrinfo.NB);
  setVSC(pcrinfo.VSC);
  setAFC(pcrinfo.AFC);
  setCTCSS(pcrinfo.CTCSS);
  setAlias(pcrinfo.alias);
  setInfo(pcrinfo.info);

  strncpy(pcrinfo.device, device, 64);
  
  pcrinfo.isPowerOn = true;
  
  return true;
}





void PcrProxy::ResetRadio(){
  QMessageBox::warning(0, "QtPCR",
                       "QtPCR has lost connection with PCR-1000 radio.\n"
                       "Re-initiating connection ...");
  InitRadio( pcrinfo.device );
}



//
// TurnOffRadio sends command to turn radio off
// and close serial port
//
bool PcrProxy::TurnOffRadio()
{
  emit sendMessage("H100\x0d\x0a",6);
  sleep( 1 );
  pcrIO.Close();
}


//
// freqParse parses freqency, mode, and filter into valid radio
// command string
// Sometimes the filter value does not valid for a given mode,
// it automatically reset filter to a valid value and send 
// filterUpdate signal
//
bool PcrProxy::freqParse(pcrfreq_t freq, modulate_t mode, filter_t filter, char *str, int *len)
{
  // check filter
  switch( mode ){
  case MODE_NFM: filter = clipped(FILTER_6kHz, filter, FILTER_50kHz); break;
  case MODE_WFM: filter = clipped(FILTER_50kHz, filter, FILTER_230kHz); break;
  case MODE_AM : filter = clipped(FILTER_3kHz, filter, FILTER_50kHz); break;
  case MODE_USB: case MODE_LSB: case MODE_CW:
    filter = clipped(FILTER_3kHz, filter, FILTER_6kHz);
    break;			
  }
  
  pcrinfo.filter = filter;
  
  sprintf(str, "K0%0.10lu%0.2X%0.2X00\x0d\x0a", freq, mode, filter);
  *len = strlen(str);
  
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::freqParse : %s\n", str);
#endif // DEBUG_VER_
  
  return true;
}




bool PcrProxy::setFreq(pcrfreq_t freq)
{
  int len;

  
  // check valid freqency range
  if((FREQ_MIN < freq) && (freq < FREQ_MAX)){

#ifdef SMART_RESETINFO
    if(fabs(freq-pcrinfo.freq) > 2*FILTER_TO_HERTZ(pcrinfo.filter)){
      bzero(pcrinfo.alias, ALIASSIZE);
      bzero(pcrinfo.info, INFOSIZE);
      emit aliasUpdate( pcrinfo.alias );
      emit infoUpdate( pcrinfo.info );
    }
#endif    

    pcrinfo.freq = freq;
    freqParse(pcrinfo.freq, pcrinfo.mode, pcrinfo.filter, buffer, &len);
    emit sendMessage(buffer, len); 	
    emit freqUpdate( freq );

    return true;

  }else
    return false;
	
}




bool PcrProxy::setMode(modulate_t mode)
{
  int len;
	
  pcrinfo.mode = mode;
  freqParse(pcrinfo.freq, pcrinfo.mode, pcrinfo.filter, buffer, &len);
  emit sendMessage(buffer, len);
  emit filterUpdate( pcrinfo.filter );
  emit modeUpdate( pcrinfo.mode );
  return true;
}




bool PcrProxy::stepUpFilter()
{
  switch( pcrinfo.filter ){
  case FILTER_3kHz : return setFilter( FILTER_6kHz   ); break;
  case FILTER_6kHz : return setFilter( FILTER_15kHz  ); break;
  case FILTER_15kHz: return setFilter( FILTER_50kHz  ); break;
  case FILTER_50kHz: return setFilter( FILTER_230kHz ); break;
  }
}



bool PcrProxy::stepDownFilter()
{
  switch( pcrinfo.filter ){
  case FILTER_230kHz: return setFilter( FILTER_50kHz  ); break;
  case FILTER_50kHz : return setFilter( FILTER_15kHz  ); break;
  case FILTER_15kHz : return setFilter( FILTER_6kHz   ); break;
  case FILTER_6kHz  : return setFilter( FILTER_3kHz   ); break;
  }
}




bool PcrProxy::setFilter(filter_t filter)
{
  int len;

#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setFilter %d\n", (int)filter);
#endif // DEBUG_VER_

  pcrinfo.filter = filter;
  freqParse(pcrinfo.freq, pcrinfo.mode, pcrinfo.filter, buffer, &len);
  emit sendMessage(buffer, len);
  emit filterUpdate( pcrinfo.filter );
  return true;
}


bool PcrProxy::setIFShift(ifshift_t shift)
{
  pcrinfo.ifShift = shift;
  sprintf(buffer, "J43%0.2X\x0d\x0a", shift);
  emit sendMessage(buffer, 7);
  emit ifShiftUpdate( shift );
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setIFShift : %s\n", buffer);
#endif // DEBUG_VER_
  return true;
}


bool PcrProxy::setVol(vol_t vol)
{
  sprintf(buffer, "J40%0.2X\x0d\x0a", vol);

  pcrinfo.vol = vol;

  emit sendMessage(buffer, 7);
  emit volUpdate( vol );
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setVol : %s\n", buffer);
#endif // DEBUG_VER_
  return true;
}



bool PcrProxy::setATT( bool val )
{
  if( val )
    sprintf(buffer, "J4701\x0d\x0a");
  else
    sprintf(buffer, "J4700\x0d\x0a");
  emit sendMessage(buffer, 7);
  emit attUpdate( val );

  pcrinfo.ATT = val;
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setATT : %s\n", buffer);
#endif // DEBUG_VER_

}


bool PcrProxy::setVSC( bool val )
{
  if( val )
    sprintf(buffer, "J5001\x0d\x0a");
  else
    sprintf(buffer, "J5000\x0d\x0a");
  emit sendMessage(buffer, 7);
  emit vscUpdate( val );

  pcrinfo.VSC = val;
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setVSC : %s\n", buffer);
#endif // DEBUG_VER_

}


bool PcrProxy::setAGC( bool val )
{
  if( val )
    sprintf(buffer, "J4501\x0d\x0a");
  else
    sprintf(buffer, "J4500\x0d\x0a");
  emit sendMessage(buffer, 7);
  emit agcUpdate( val );

  pcrinfo.AGC = val;
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setAGC : %s\n", buffer);
#endif // DEBUG_VER_

}


bool PcrProxy::setAFC( bool val )
{
  pcrinfo.AFC = val;
  emit afcUpdate( val );
}


bool PcrProxy::setNB( bool val )
{
  if( val )
    sprintf(buffer, "J4601\x0d\x0a");
  else
    sprintf(buffer, "J4600\x0d\x0a");
  emit sendMessage(buffer, 7);
  emit nbUpdate( val );

  pcrinfo.NB = val;
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setNB : %s\n", buffer);
#endif // DEBUG_VER_

}



bool PcrProxy::setSquelch( squelch_t squ )
{
  sprintf(buffer, "J41%0.2X\x0d\x0a", squ);
  emit sendMessage(buffer, 7);
  emit squelUpdate( squ );

  pcrinfo.squelch = squ;
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setSquelch : %s\n", buffer);
#endif // DEBUG_VER_

}


// sending tone = 0xFF reset to previous CTCSS
bool PcrProxy::setCTCSS( unsigned char tone )
{
  // check range
  if( tone > 0x33 ){

    if(pcrinfo.previousCTCSS)
      tone = pcrinfo.previousCTCSS;
    else
      tone = 0x01;
  }
  pcrinfo.previousCTCSS = pcrinfo.CTCSS;
  pcrinfo.CTCSS = tone;

  sprintf(buffer, "J51%0.2X\x0d\x0a", tone);
  emit sendMessage(buffer, 7);
  emit ctcssUpdate( tone );

  pcrinfo.CTCSS = tone;
#ifdef DEBUG_VER_
  fprintf(stderr, "PcrProxy::setCTCSS : %s\n", buffer);
#endif // DEBUG_VER_

}



bool PcrProxy::setTS( pcrfreq_t ts )
{
  pcrinfo.ts = ts;

  if(pcrinfo.isBandScopeOn){
    scopeOn( true );
  }

  emit tsUpdate( ts );
}


bool PcrProxy::stepUpTS()
{
  // KISS (Keep It Simple and Stupid) algorithm
  switch( pcrinfo.ts ){
    // Dave, VK5DG, added low freqency steps 
  case TS_10Hz   : setTS( TS_25Hz ); break;
  case TS_25Hz   : setTS( TS_50Hz ); break;
  case TS_50Hz   : setTS( TS_100Hz ); break;
  case TS_100Hz  : setTS( TS_250Hz ); break;
  case TS_250Hz  : setTS( TS_500Hz ); break;
  case TS_500Hz  : setTS( TS_1000Hz ); break;
  case TS_1000Hz : setTS( TS_2500Hz ); break;
  case TS_2500Hz : setTS( TS_5000Hz ); break;
  case TS_5000Hz : setTS( TS_10000Hz ); break;
  case TS_10000Hz: setTS( TS_25000Hz ); break;
  }
}


bool PcrProxy::stepDownTS()
{
  // KISS (Keep It Simple and Stupid) algorithm
    switch( pcrinfo.ts ){
      // Dave, VK5DG, added low freqency steps
    case TS_25000Hz: setTS( TS_10000Hz ); break;
    case TS_10000Hz: setTS( TS_5000Hz ); break;
    case TS_5000Hz : setTS( TS_2500Hz ); break;
    case TS_2500Hz : setTS( TS_1000Hz ); break;
    case TS_1000Hz : setTS( TS_500Hz ); break;
    case TS_500Hz  : setTS( TS_250Hz ); break;
    case TS_250Hz  : setTS( TS_100Hz ); break;
    case TS_100Hz  : setTS( TS_50Hz ); break;
    case TS_50Hz   : setTS( TS_25Hz ); break;
    case TS_25Hz   : setTS( TS_10Hz ); break;
    }
}

bool PcrProxy::stepUpFreq()
{
  pcrfreq_t newFreq;

  newFreq = pcrinfo.freq + pcrinfo.ts;
  if(setFreq( newFreq )){
    pcrinfo.freq = newFreq;
    return true;
  }else
    return false;
}




bool PcrProxy::stepDownFreq()
{
  pcrfreq_t newFreq;

  newFreq = pcrinfo.freq - pcrinfo.ts;
  if(setFreq( newFreq )){
    pcrinfo.freq = newFreq;
    return true;
  }else
    return false;
}



bool PcrProxy::stepFreq(int step)
{
  pcrfreq_t newFreq;

  newFreq = pcrinfo.freq + step*pcrinfo.ts;
  if(setFreq( newFreq )){
    pcrinfo.freq = newFreq;
    return true;
  }else
    return false;
}	



//
// charHexToDex convert a single HEX character '0'..'F'
// into a decival value
//
unsigned char PcrProxy::charHexToDec(char digit)
{
  if(('A' <= digit) && (digit <= 'F'))
    return digit-'A'+10;
  else if(('1' <= digit) && (digit <= '9'))
    return digit-'1'+1;
  else
    return 0;
}



//
// radioMessageSlot waits for messages from PcrIO
//
void PcrProxy::radioMessageSlot(const char *mesg, int len)
{
  int index;
  int i;
  char digit;
  char buffer[PCRIO_BUFFERSIZE];
  unsigned char val;

  memcpy( buffer, mesg, len );

  // kill Initradio Timer
  timerInitRadio.stop();

  // under auto update mode, there are two types of responses
  // thoses begin with 'N' for bandscope and 'I' for normal
  // inquiries


  // bandscope
  if(buffer[0] == 'N'){

    digit = buffer[3];
    index = (int)charHexToDec(digit)*16;

    for(i = 0; i < 16; i++){

      digit = buffer[i*2 + 5];
      val = charHexToDec(digit)*16;
      digit = buffer[i*2 + 6];
      val += charHexToDec(digit);
      pcrinfo.bandScope[index+i] = val;
    }

    emit bandScopeUpdate(pcrinfo.bandScope, pcrinfo.bandScopeNsamples);
  }

  // normal inquires 
  if(buffer[0] == 'I'){
    switch(buffer[1]){

      // squelch ON/ OFF
    case '0':
      digit = buffer[3];
      if(digit == '4')
	emit squelchOnUpdate( false );
      else 
	emit squelchOnUpdate( true );
      break;

      // sig strength
    case '1':
      digit = buffer[2];
      val = charHexToDec(digit)*16;	
      digit = buffer[3];
      val += charHexToDec(digit);
      pcrinfo.sigStrength = val;

      emit sigStrengthUpdate( val );
      break;

      // AFC center 
    case '2':
      digit = buffer[2];
      val = charHexToDec(digit)*16;
      digit = buffer[3];
      val += charHexToDec(digit);
      
      pcrinfo.sigCenter = val;
      emit sigCenterUpdate( val );

      // control frequency
      if(pcrinfo.AFC){
	switch( val ){
	case AFC_LOW: stepDownFreq(); break;
	case AFC_HIGH: stepUpFreq(); break;
	}
      }
      break;
    }
  }
}


bool PcrProxy::scopeOn( bool flag)
{
  unsigned char rate;
  unsigned char nsamples;

  if(flag){

    nsamples = pcrinfo.bandScopeNsamples;
    rate = 0x05;
    
    // parse band scope command
    sprintf(buffer, "ME00001%0.2X%0.2X0100%0.6d\x0d\x0a", 
	    nsamples, 
	    rate,  
	    pcrinfo.ts );
    
    emit sendMessage( buffer, 23);
    emit scopeOnUpdate( true );

    pcrinfo.isBandScopeOn = true;
    
#ifdef DEBUG_VER_
    fprintf( stderr, "PcrProxy::scopeOn - emit %s \n", buffer );
#endif	// DEBUG_VER_
  }else{
    sprintf( buffer, "ME0000114050000010000\x0d\x0a" );
    emit sendMessage( buffer , 23);
    emit scopeOnUpdate( false );

    pcrinfo.isBandScopeOn = false;
#ifdef DEBUG_VER_
    fprintf( stderr, "PcrProxy::scopeOn - emit %s \n", buffer );
#endif  // DEBUG_VER_
    
  }

}

bool PcrProxy::scopeWide()
{
  if(pcrinfo.isBandScopeOn && (int)pcrinfo.bandScopeNsamples*2 < 255){
    pcrinfo.bandScopeNsamples  = pcrinfo.bandScopeNsamples*2;	
    scopeOn( true );
  }
}


bool PcrProxy::scopeNarrow()
{
  if(pcrinfo.isBandScopeOn && (int)pcrinfo.bandScopeNsamples/2 > 8){
    pcrinfo.bandScopeNsamples  = pcrinfo.bandScopeNsamples/2;
    scopeOn( true );
  }
}


bool PcrProxy::setAlias( const char * str )
{
  strncpy(pcrinfo.alias, str, ALIASSIZE);
  emit aliasUpdate( pcrinfo.alias );
}

bool PcrProxy::setInfo( const char * str )
{
  strncpy(pcrinfo.info, str, INFOSIZE);
  emit infoUpdate( pcrinfo.info );
}

bool PcrProxy::getBookmark(struct bookmark_t *bm)
{
  // copy mem directly to speed up, but need to
  // be carefull every time we change bookmark_t
  memcpy(bm, &pcrinfo, sizeof(struct bookmark_t));
}


bool PcrProxy::gotoBookmark(struct bookmark_t *bm)
{
  int len;

  // set auxilary mode first  
  setSquelch(bm->squelch);
  setATT(bm->ATT);
  setAGC(bm->AGC);
  setNB(bm->NB);
  setVSC(bm->VSC);
  setAFC(bm->AFC);
  setCTCSS(bm->CTCSS);
  setIFShift(bm->ifshift);
  setTS(bm->ts);
  
  // this should be hard-coded
  pcrinfo.freq   = bm->freq;
  pcrinfo.mode   = bm->mode;
  pcrinfo.filter = bm->filter;
  freqParse(pcrinfo.freq, pcrinfo.mode, pcrinfo.filter, buffer, &len);

  emit sendMessage(buffer, len);
  emit freqUpdate( pcrinfo.freq );
  emit modeUpdate( pcrinfo.mode );
  emit filterUpdate( pcrinfo.filter );

  setVol(bm->vol);
  scopeOn( bm->scope );
  
  setAlias( bm->alias );
  setInfo( bm->info );
}

