#include <qgroupbox.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qmessagebox.h>
#include <stdio.h>
#include "scandialog.h"

SignalLine::SignalLine( QCanvas * canvas )
  : QCanvasLine( canvas )
{
  freq = DEFAULT_FREQ;
  sig  = 0;
}

int SignalLine::rtti() const{ return SignalLineID; }
pcrfreq_t SignalLine::getFrequency(){ return freq; }
unsigned char SignalLine::getSig(){ return sig; }
void SignalLine::setFrequency(pcrfreq_t freq){ this->freq = freq; }
void SignalLine::setSig( unsigned char sig){ this->sig = sig; }

ScanScreen::ScanScreen( QCanvas * viewing, QWidget * parent, const char * name, WFlags f)
  : QCanvasView( viewing, parent, name, f)
{
    clear();
}


void ScanScreen::clear()
{
    QCanvasItemList list = canvas()->allItems();
    QCanvasItemList::Iterator it = list.begin();
    for (; it != list.end(); ++it) {
    if ( *it )
        delete *it;
    }
}


void ScanScreen::contentsMouseDoubleClickEvent ( QMouseEvent * e )
{
  QCanvasItemList l=canvas()->collisions(e->pos());
  QCanvasItemList::Iterator it = l.begin();
  if( it != l.end()){
    if((*it)->rtti() == SignalLineID){
      SignalLine *item = (SignalLine *)(*it);
      emit setFreq( item->getFrequency() );
    }
  }
}


ScanDialog::ScanDialog( QWidget * parent, const char * name )
  : QDialog( parent, name )
{
  QGroupBox *GBoxPtr;
  QGridLayout *MainGrid;

  // create main grid 
  MainGrid = new QGridLayout( this, 1, 2 );
  GBoxPtr  = new QGroupBox( 7, Qt::Vertical, "Scan", this );

  GBoxPtr->setMargin(10);

  // create buttons
  start = new QPushButton("Start", GBoxPtr);
  pause = new QPushButton("Pause", GBoxPtr);
  stop  = new QPushButton("Stop",  GBoxPtr);

  connect( start, SIGNAL(clicked()),
           this, SLOT(StartSlot()));

  connect( stop, SIGNAL(clicked()),
           this, SLOT(StopSlot()));

  connect( pause, SIGNAL(clicked()),
           this, SLOT(PauseSlot()));

  stop->setEnabled( false );
  pause->setEnabled( false );


  stop->setFocusPolicy( NoFocus );
  start->setFocusPolicy( NoFocus );
  // pause->setFocusPolicy( NoFocus );

  // create edit box
  startFreqEdit  = new QLineEdit( GBoxPtr );
  stopFreqEdit   = new QLineEdit( GBoxPtr );

  startFreqEdit->setMaxLength(10);
  startFreqEdit->setAlignment( Qt::AlignRight);
  
  stopFreqEdit->setMaxLength(10);
  stopFreqEdit->setAlignment( Qt::AlignRight);
 
  // create mode combobox
  mode = new QComboBox( GBoxPtr );
  filter = new QComboBox( GBoxPtr );

  mode->insertItem("NFM");
  mode->insertItem("AM");
  mode->insertItem("USB");
  mode->insertItem("LSB");
  mode->insertItem("WFM");

  filter->insertItem("3kHz");
  filter->insertItem("6kHz");
  filter->insertItem("15kHz");
  filter->insertItem("50kHz");
  filter->insertItem("230kHz");

  mode->setFocusPolicy( NoFocus );
  filter->setFocusPolicy( NoFocus );

  // create Canvas
  canvas = new QCanvas(600, 256);
  canvas->setBackgroundColor( Qt::black );
  screen = new ScanScreen( canvas, this );

  MainGrid->addWidget( screen, 0, 0);
  MainGrid->addWidget( GBoxPtr, 0, 1);  

  scan = new Scan();

  connect( screen, SIGNAL(setFreq(pcrfreq_t)),
           this,   SIGNAL(setFreq(pcrfreq_t)));

  connect( scan, SIGNAL(scanUpdate(pcrfreq_t, unsigned char)),
           this, SLOT(scanUpdate(pcrfreq_t, unsigned char)));

  connect( scan, SIGNAL(scanBegun()),
           this, SLOT(scanBegun()));

  connect( scan, SIGNAL(scanEnded()),
           this, SLOT(scanEnded()));

  connect( scan, SIGNAL(setFreq(pcrfreq_t)),
           this, SIGNAL(setFreq(pcrfreq_t)));

  connect( scan, SIGNAL(setMode(modulate_t)),
           this, SIGNAL(setMode(modulate_t)));

  connect( scan, SIGNAL(setFilter(filter_t)),
           this, SIGNAL(setFilter(filter_t)));

}

ScanDialog::~ScanDialog(){
  screen->clear();
  delete screen;
  delete canvas;
  delete scan;
  delete pause;
  delete start;
  delete stop;
}

void ScanDialog::sigStrengthUpdate( unsigned char sig)
{
  scan->sigStrengthUpdate( sig);
}

void ScanDialog::squelchOnUpdate( bool flag)
{
  scan->squelchOnUpdate( flag );
}


void ScanDialog::StartSlot()
{
   bool ok;
   modulate_t ch_mode;
   filter_t ch_filter;

   if(mode->currentText().contains("NFM")) ch_mode = MODE_NFM; else
   if(mode->currentText().contains("WFM")) ch_mode = MODE_WFM; else
   if(mode->currentText().contains("AM")) ch_mode = MODE_AM;   else
   if(mode->currentText().contains("LSB")) ch_mode = MODE_LSB; else
   if(mode->currentText().contains("USB")) ch_mode = MODE_USB; else{
      QMessageBox::warning(this, "QtPCR", "Cannot determine mode");
      return;
   }

   if(filter->currentText().contains("3kHz")) ch_filter = FILTER_3kHz; else
   if(filter->currentText().contains("6kHz")) ch_filter = FILTER_6kHz; else
   if(filter->currentText().contains("15kHz")) ch_filter = FILTER_15kHz; else
   if(filter->currentText().contains("50kHz")) ch_filter = FILTER_50kHz; else
   if(filter->currentText().contains("230kHz")) ch_filter = FILTER_230kHz; else {
      QMessageBox::warning(this, "QtPCR", "Cannot determine filter");
      return;
   }

   endFreq  = stopFreqEdit->text().toULong(&ok);
   if(!ok){
      QMessageBox::warning(this,"QtPCR","Cannot determine ending frequency");
      return;
   }
 
   beginFreq = startFreqEdit->text().toULong(&ok);
   if(!ok){
      QMessageBox::warning(this, "QtPCR","Cannot determine starting frequency");
      return;
   }

   if(beginFreq < FREQ_MIN || beginFreq > FREQ_MAX){
      QMessageBox::warning(this, "QtPCR","Starting frequency out of range");
      return;
   }

   if(endFreq < FREQ_MIN || endFreq > FREQ_MAX){
      QMessageBox::warning(this, "QtPCR","Ending frequency out of range");
      return;
   }

   if(beginFreq > endFreq ){
      QMessageBox::warning(this, "QtPCR","Begin frequency cannot be larger than ending frequency");
      return;
   }

   freq = beginFreq;

   scan->SetFrequency(beginFreq, endFreq);
   scan->SetChannel( ch_mode, ch_filter);
   scan->SetTimeDelay(100);

   stopFreqEdit->setEnabled( false );
   startFreqEdit->setEnabled( false );
   start->setEnabled( false );
   mode->setEnabled( false );
   filter->setEnabled( false );
   stop->setEnabled( true );
   pause->setEnabled( true );
   screen->clear();
   scan->Start();
   x = 1;

   // draw horizontal line
   QCanvasLine *l = new QCanvasLine(canvas);
    l->setPoints( 0, 255, canvas->width(), 255);
    l->setZ(+10);
    l->setPen( QPen( Qt::red, 1) );
    l->show();
}

void ScanDialog::StopSlot()
{
   scan->Stop();

   stopFreqEdit->setEnabled( true );
   startFreqEdit->setEnabled( true );
   start->setEnabled( true );
   mode->setEnabled( true );
   filter->setEnabled( true );
   stop->setEnabled( false );
   pause->setEnabled( false );
}

void ScanDialog::PauseSlot()
{
   QMessageBox::warning(this, "QtPCR",
                              "Not yet implemented");
}


void ScanDialog::scanBegun(){}
void ScanDialog::scanEnded(){
   stopFreqEdit->setEnabled( true );
   startFreqEdit->setEnabled( true );
   start->setEnabled( true );
   mode->setEnabled( true );
   filter->setEnabled( true );
   stop->setEnabled( false );
   pause->setEnabled( false );
}

void ScanDialog::scanUpdate( pcrfreq_t freq, unsigned char sig)
{
  SignalLine *i; 

  // low signal is a waste of memory
  if(sig > 2){ 
    i = new SignalLine(canvas);
    i->setPoints(x, 256, x, 256-sig);
    i->setPen( QPen( QColor(0,sig,0), 1) );
    i->show();
    i->setFrequency( freq );
    i->setSig( sig );
  }
  x++;
  this->freq = freq;

  // expand canvas as nesscessary
  if(x > canvas->width()){

   // draw horizontal line
   QCanvasLine *l = new QCanvasLine(canvas);
    l->setPoints( canvas->width(), 255, canvas->width()+600, 255);
    l->setZ(+10);
    l->setPen( QPen( Qt::red, 1) );

    canvas->resize(canvas->width()+600, 256);
    l->show();
  }

  // mark some frequency
  if(x%100 == 0){
    QString str;
    QCanvasText *t = new QCanvasText( canvas );
 
    // draw Text
    if( freq >= 1000000000)
       str.sprintf("%6.3fGHz",freq/1000000000.0);
    else if( freq >= 1000000) 
       str.sprintf("%6.3fMHz",freq/1000000.0);
    else 
       str.sprintf("%6.3fkHz",freq/1000.0);

    t->setX(x+7);
    t->setY(30);
    t->setZ(10);
    t->setText( str );
    t->setColor( Qt::red );
    t->show();

    // draw Line
    QCanvasLine *l = new QCanvasLine( canvas );
    l->setPoints( x, 256, x, 35);
    l->setZ(-10);
    l->setPen( QPen( Qt::red, 1) );
    l->show();

    l = new QCanvasLine( canvas );
    l->setPoints( x, 35, x+5, 35);
    l->setPen( QPen( Qt::red, 1));
    l->setZ(-10);
    l->show();
  }

  canvas->update();
}
